#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <EEPROM.h>
#include <math.h>

//LM75 defines
#define SensorAdresse 0x48  // Baseadress first LM75
#define K03 0x3A            // Expander D/A Converter
#define K16 0x38            // Expander LCD
#define TEMP 0              // Temp register
#define UNTEN 2             // Lower threshold 
#define OBEN 3              // Upper treshold
#define MAXTEMP 41.00       // Maximum value for the OS-output
#define R01OS 6             // Room 01 OS to arduino
#define R02OS 5             // Room 02 OS to arduino

#define InitEEPROM 0
#define BetriebsLEDEEPROM 1
#define BLED 7

#define S1 digitalRead(2)
#define S2 digitalRead(3)
#define S3 digitalRead(4)

LiquidCrystal_I2C lcd(K16, 16, 4);
#define LCDOFFSET -4

// Functions
// Output CheckCodes and ErrorCodes
// Meaning of the codes can be taken from the table.
void statusCode(const String &);
void checkCode(const String &);
int statusLED(boolean);

// LCD Output and waiting of S2
void outputLCD(const String &, const String &);
void readS2(void);

// Read the initialization Bit EEPROM
int readInitBit(void);

// Initialization of the climate control
int InitClimaticControll(void);
void InitRoom(int roomID);
void initVINVNE(int opt);

// Output expander (Input of D/A Converter)
void sendDigitalValue(byte ADDR, byte value);

// LM75
double get_LM75_temperature(int device, int regx);
void set_LM75_config(int device, byte value);
byte get_LM75_config(int device);
void set_LM75_schaltwert(int device, byte regx, double grad);

// Measurements ...
void goToDA(void);
void goToI2C(void);
void goToInitK18(void);
void checkRooms(int);

// global variables
char dataString[7]; // read temperature processed as a string: (-xx)x.x
double tmpRoom1, tmpRoom2;


void setup() {
  // put your setup code here, to run once:
  int statusInit = 0;
  Wire.begin();

  Serial.begin(9600);
  Serial.println("Programmübertragung abgeschlossen!");
  Serial.println("Prüfung Sommer 20");
  Serial.println("Sie können den Arduino abziehen!");
  Serial.end();

  analogReference(EXTERNAL);
  for (int i = 2; i < 6; i++) pinMode(i, INPUT);
  for (int i = 7; i < 12; i++) {
    pinMode(i, OUTPUT); digitalWrite(i, LOW);
  }
  pinMode(R01OS, INPUT);
  pinMode(R02OS, INPUT);
  lcd.init();                      // initialize the lcd
  lcd.init();
  outputLCD(" Willkommen zur ", "    Pruefung    ");
  delay(1500);
  sendDigitalValue(K03, 0xFF); //OPV Set output voltage to maximum
  statusInit = readInitBit();
  outputLCD("  Booting  0 %   ", "Init PORTs, I2C ");
  for (int i = 1; i < 4; i++, delay(500)) //Delay to temper the sensors
  {
    lcd.setCursor(11, 1);
    lcd.print(i);
  }
  outputLCD("  Booting 5 %   ", "                ");
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("    | INIT |    ");

  for (int i = 10; i < 81; i++, delay(100))  //Delay to temper the sensors
  {

    lcd.setCursor(10, 1);
    lcd.print(int(i * 1.25));
    lcd.print(" %  ");
    if (!S2)
    {
      while (!S2) delay(100);
      statusInit = 2;
      i = 81;
    }
  }

  // Set thresholds
  // Room 01
  tmpRoom1 = get_LM75_temperature(1, TEMP);
  set_LM75_schaltwert(1, UNTEN, MAXTEMP - 3.00);
  set_LM75_schaltwert(1, OBEN, MAXTEMP);

  // Room 02
  tmpRoom2 = get_LM75_temperature(3, TEMP);
  set_LM75_schaltwert(3, UNTEN, MAXTEMP - 3.00);
  set_LM75_schaltwert(3, OBEN, MAXTEMP);

  // Set min bargraph temperature
  tmpRoom1 = (tmpRoom1 > 32 ? 25 : tmpRoom1); //get_LM75_temperature(1, TEMP);
  tmpRoom2 = (tmpRoom2 > 32 ? 25 : tmpRoom2); //get_LM75_temperature(3, TEMP);
  outputLCD("                ", "                ");
  if (!statusInit || statusInit == 2)
  {
    do  {
      checkCode("INIT");
      lcd.setCursor(LCDOFFSET, 3);
      lcd.print("    | NEXT |    ");
      readS2();
      statusInit = InitClimaticControll();
    }
    while (statusInit != 1);
    EEPROM.write(InitEEPROM, statusInit);
    outputLCD("Inbetriebnahme  ", "abgeschlossen!  ");
    delay(2000);
  }/**/
}

void loop() {
  // put your main code here, to run repeatedly:
  statusLED(true);
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("    | MENU |    ");
  checkRoom(1);
  checkRoom(2);
}

void goToMenu() {
  int menu = 0;
  digitalWrite(11, LOW);
  digitalWrite(9, LOW);
  digitalWrite(10, LOW);
  digitalWrite(8, LOW);
  while (!S2) delay(100);
  outputLCD("                ", "                ");
  while (1)
  {
    lcd.setCursor(LCDOFFSET, 3);
    lcd.print("BACK |START| -> ");
    do {
      if (!S3)
      {
        while (!S3) delay(100);
        menu++;
        menu = (menu > 3 ? 0 : menu);
      }
      else if (!S1)
      {
        while (S1) delay(100);
        return;
      }
      lcd.setCursor(0, 1);
      switch (menu)
      {
        case 0: lcd.print("Aufg. 2 Bl.4    "); break;
        case 1: lcd.print("D/A - Messung   "); break;
        case 2: lcd.print("I2C - Messung   "); break;
        case 3: lcd.print("Room Information"); break;
      }

    }
    while (S2);
    while (!S2) delay(100);
    switch (menu)
    {
      case 0: goToInitK18(); break;
      case 1: goToDA(); break;
      case 2: goToI2C(); break;
      case 3: goToRoom(); break;
    }
    checkCode("    ");
    outputLCD("                ", "                ");
  }
  return;
}

void goToDA() {
  int davalue = 0;
  statusLED(true);
  checkCode("MESS");
  outputLCD("D/A-Messung ist ", "aktiv!          ");
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print(" <  |  END |  > ");
  do {
    if (!S3) {
      while (!S3) delay(100);
      davalue++;
    }
    else if (!S1) {
      while (!S1) delay(100);
      davalue--;
    }
    if (davalue < 0) davalue = 0;
    if (davalue > 15) davalue = 15;
    statusLED(true);
    lcd.setCursor(8, 2);
    lcd.print((davalue & 0x08 ? 1 : 0));
    lcd.print((davalue & 0x04 ? 1 : 0));
    lcd.print((davalue & 0x02 ? 1 : 0));
    lcd.print((davalue & 0x01 ? 1 : 0));
    sendDigitalValue(K03, davalue);
    delay(10);
  } while (S2);
  while (!S2) delay(100);
  outputLCD("D/A-Messung ist ", "abgeschlossen!  ");
  delay(1250);
}

void goToI2C() {
  statusLED(true);
  checkCode("MESS");
  outputLCD("I2C-Messung ist ", "aktiv!          ");
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("             END");
  do {
    sendDigitalValue(K03, 0x74);
    delay(10);
  } while (S3);
  while (!S3) delay(100);
  outputLCD("I2C-Messung ist ", "abgeschlossen!  ");
  delay(1250);
}

void goToInitK18() {
  statusLED(true);
  checkCode("INIT");
  outputLCD("Aufgabe 2 Bl. 4 ", "Inbetriebnahme  ");
  sendDigitalValue(K03, 0x08);
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("    | NEXT |    ");
  readS2();
  sendDigitalValue(K03, 0x00);
  outputLCD("Inbetriebnahme  ", "abgeschlossen!  ");
  delay(1250);
}

void goToRoom()
{
  int roomID = 1;
  statusCode("R0");
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("BACK |  <  |  > ");
  outputLCD("OS-ON:        \337C",
            "OS-OFF:       \337C");
  do {
    if (!S3) {
      while (!S3) delay(100);
      roomID++;
    }
    else if (!S2) {
      while (!S2) delay(100);
      roomID--;
    }
    if (roomID < 1) roomID = 2;
    if (roomID > 2) roomID = 1;
    lcd.setCursor(14, 0);
    lcd.print(roomID);
    lcd.setCursor(0, 0);
    lcd.print(get_LM75_temperature((roomID == 1 ? 1 : 3), TEMP));
    lcd.print("\337C");
    lcd.setCursor(9, 1);
    lcd.print(get_LM75_temperature((roomID == 1 ? 1 : 3), OBEN));
    lcd.setCursor(LCDOFFSET + 9, 2);
    lcd.print(get_LM75_temperature((roomID == 1 ? 1 : 3), UNTEN));
    delay(10);
  } while (S1);
  while (!S1) delay(100);
  checkCode("       ");
  statusCode("    ");
  return;
}

int readInitBit() {
  int readEEPROM;
  readEEPROM = EEPROM.read(InitEEPROM);
  return ((readEEPROM == 0 || readEEPROM == 1) ? readEEPROM : 0);
}

void outputLCD(const String &n1, const String &n2) {
  lcd.setCursor(0, 1);
  lcd.print(n1);
  lcd.setCursor(LCDOFFSET, 2);
  lcd.print(n2);
  return;
}

void readS2() {
  while (S2);
  while (!S2) delay(100);
  return;
}

void checkCode(const String &n1)  {
  lcd.setCursor(0, 0);
  lcd.print(n1);
}

void statusCode(const String &n1)  {
  lcd.setCursor(12, 0);
  lcd.print(n1);
  return;
}

int statusLED(boolean set)
{
  int LEDonoff = EEPROM.read(BetriebsLEDEEPROM);
  digitalWrite(BLED, (set ? LEDonoff : !LEDonoff));
}

void checkRoom(int roomID)
{
  double roomTemp[] = { (roomID == 1 ? tmpRoom1 : tmpRoom2) + 2.00 , (roomID == 1 ? tmpRoom1 : tmpRoom2) + 3.50 }; //dynamic
  double readTemp = 0.00;
  int davalue[2] = { 0, 0 };
  lcd.setCursor((roomID == 1 ? 0 : LCDOFFSET), roomID);
  lcd.print("RO" + (String)roomID + "-Temp:     \337C");
  lcd.setCursor((roomID == 1 ? 0 : LCDOFFSET) + 9, roomID);
  readTemp = get_LM75_temperature((roomID == 1 ? 1 : 3), TEMP);
  lcd.print(readTemp);

  // Convert temp to 4 Bit
  davalue[0] = map((int)get_LM75_temperature(1, TEMP), (int)20, (int)(MAXTEMP < (tmpRoom1 + 15.00) ? MAXTEMP : (tmpRoom1 + 15.00)), 0, 15);       // lower 4 Bit (Room 01)
  davalue[1] = (map((int)get_LM75_temperature(3, TEMP), (int)20, (int)(MAXTEMP < (tmpRoom2 + 15.00) ? MAXTEMP : (tmpRoom2 + 15.00)), 0, 15) << 4); // higher 4 Bit (Room 02)
  sendDigitalValue(K03, ((davalue[1] & 0xF0) | (davalue[0] & 0x0F)));
  if (readTemp >= roomTemp[(roomID - 1)])
  {
    digitalWrite((roomID == 1 ? 11 : 9), HIGH);
    digitalWrite((roomID == 1 ? 10 : 8), LOW);
  }
  else
  {
    digitalWrite((roomID == 1 ? 11 : 9), LOW);
    digitalWrite((roomID == 1 ? 10 : 8), HIGH);
  }
  for (int i = 0; i < 50; i++, delay(10))
  {

    while (!digitalRead(R01OS) || !digitalRead(R02OS))
    {
      statusCode("ER0X");
      statusLED(false);
      lcd.setCursor(LCDOFFSET, 3);
      lcd.print("    |      |    ");
      digitalWrite(11, !digitalRead(R01OS) ? HIGH : LOW);
      digitalWrite(9, !digitalRead(R02OS) ? HIGH : LOW);
      digitalWrite(10, LOW);
      digitalWrite(8, LOW);
      lcd.setCursor(9, 1);
      readTemp = get_LM75_temperature(1, TEMP);
      lcd.print(readTemp);
      lcd.setCursor(LCDOFFSET + 9, 2);
      readTemp = get_LM75_temperature(3, TEMP);
      lcd.print(readTemp);
      delay(500);
    }
    if (!S2) goToMenu();
    checkCode("    ");
    statusCode("    ");
  }
  return;
}

int InitClimaticControll() {
  int taster = 0;
  int statusTaster = 0;
  checkCode("ILED");
  digitalWrite(BLED, LOW);
  outputLCD("      Init      ", "  Betriebs-LED  ");
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print(" rt |  NA  | gn ");
  do {
    statusCode("    ");
    if (!S1)
    {
      while (S1) delay(100); statusTaster = 1; checkCode("X013"); EEPROM.write(BetriebsLEDEEPROM, 1);
    }
    else if (!S2)
    {
      while (S2) delay(100); statusCode("ELED"); EEPROM.write(InitEEPROM, 0); outputLCD("  Betriebs-LED  ", "    pruefen!    ");   lcd.setCursor(LCDOFFSET, 3); lcd.print("System shutdown!"); while (1);
    }
    else if (!S3)
    {
      while (S3) delay(100); statusTaster = 1; checkCode("X012"); EEPROM.write(BetriebsLEDEEPROM, 0);
    }
  } while (!statusTaster);

  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("    | NEXT |    ");
  readS2();

  InitRoom(1);
  InitRoom(2);
  checkCode("    ");
  statusCode("    ");
  outputLCD("  R01 und R02   ", " initialisiert. ");
  delay(1500);
  initVINVNE(0);
  initVINVNE(1);
  checkCode("    ");
  statusCode("    ");
  return 1;
}

void InitRoom(int roomID) {
  int errorBit;
  double tempOLD, tempNEW, tempOSon, tempOSoff;
  statusLED(false);
  checkCode("ITE"); lcd.print(roomID);
  lcd.setCursor(0, 1);
  lcd.print("R0" + (String)roomID + "             ");
  lcd.setCursor(LCDOFFSET, 2);
  lcd.print(" initialisieren");
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("    | NEXT |    ");
  readS2();
  if (!S3) {
    while (S3) delay(100);
    outputLCD("Abbruch Raum    ", "Initialisierung");
    delay(1500);
    return;
  }
  outputLCD(" K\xF5hlung aktiv  ", "                ");
  digitalWrite((roomID == 1 ? 11 : 9), HIGH);
  readS2();

  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("    |      |    ");
  digitalWrite((roomID == 1 ? 11 : 9), LOW);
  checkCode("IH0"); lcd.print(roomID);
  outputLCD(" Heizung aktiv  ", "                ");

  tempOSon = get_LM75_temperature((roomID == 1 ? 1 : 3), OBEN);
  tempOSoff = get_LM75_temperature((roomID == 1 ? 1 : 3), UNTEN);

  set_LM75_schaltwert((roomID == 1 ? 1 : 3), OBEN, 10.00);
  set_LM75_schaltwert((roomID == 1 ? 1 : 3), UNTEN, 10.00);
  errorBit = 1;
  tempOLD = get_LM75_temperature((roomID == 1 ? 1 : 3), TEMP);
  digitalWrite((roomID == 1 ? 10 : 8), HIGH);
  lcd.setCursor(LCDOFFSET, 2);
  for (int i = 0; i < 16; i++, delay(375))
  {
    lcd.print("-");
  }
  //outputLCD( "                ", "                ");
  lcd.setCursor(LCDOFFSET, 2);

  set_LM75_schaltwert((roomID == 1 ? 1 : 3), OBEN, tempOSon);
  set_LM75_schaltwert((roomID == 1 ? 1 : 3), UNTEN, tempOSoff);

  for (int i = 0; i < 6 && errorBit; i++, delay(1500))
  {
    tempNEW = get_LM75_temperature((roomID == 1 ? 1 : 3), TEMP) - 0.5;

    if (tempNEW > tempOLD) errorBit = 0;
    lcd.print("-");
  }



  digitalWrite((roomID == 1 ? 10 : 8), LOW);

  if (errorBit)
  {
    statusCode("ETH"); lcd.print(roomID); statusLED(false);
    lcd.setCursor(LCDOFFSET, 3);
    lcd.print("System shutdown!");
    delay(3000);
    outputLCD(" Heizungsstrang ", "  LM75 pruefen!  ");
    EEPROM.write(InitEEPROM, 0);
    while (1);
  }
  outputLCD(" Temperatur ist ", " steigend...    ");

  statusLED(true);
  lcd.setCursor(LCDOFFSET, 3);
  lcd.print("    | NEXT |    ");
  readS2();
  return;
}

void initVINVNE(int opt)  {
  statusLED(false);
  checkCode("    ");
  statusCode("    ");
  if (!opt)
  {
    checkCode("PVIN");
    outputLCD("      VIN       ", "Initialisierung ");
  }
  else
  {
    checkCode("PVNE");
    outputLCD("      VNE       ", "Initialisierung ");
  }
  long int analogwert = 0;
  analogwert = (!opt ? analogRead(A0) : analogRead(A1));
  readS2();
  outputLCD("                ", "Messung beendet!");
  if (!opt)
  {
    if (analogwert >= 820 & analogwert <= 900)
    {
      statusLED(true);
      checkCode("X102");
    }
    else
    {
      checkCode("X103");
      statusCode("EVIN");
    }
  }
  else
  {
    if (analogwert >= 280 & analogwert <= 320)
    {
      statusLED(true);
      checkCode("X202");
    }
    else
    {
      checkCode("X203");
      statusCode("EVNE");
    }
  }
  readS2();
  statusLED(true);
}

void sendDigitalValue(byte ADDR, byte value)
{
  Wire.beginTransmission(ADDR);
  Wire.write(value);
  Wire.endTransmission();
}

// LM75 Temperatur auslesen. Device = 0-7, regx = TEMP, OBEN, UNTEN (Registerauswahl)
double get_LM75_temperature(int device, int regx)
{
  int8_t msb;
  int8_t lsb;
  int8_t msb1;
  Wire.beginTransmission(SensorAdresse + device);
  Wire.write(regx);
  Wire.endTransmission();
  Wire.beginTransmission(SensorAdresse + device);
  Wire.requestFrom(SensorAdresse + device, 2);
  if (Wire.available()) {
    msb1 = Wire.read();
    msb = msb1 << 1; // Vorzeichenbit entfernen, verbliebener Wert ist nun doppelt so groß
    lsb = Wire.read();
  }
  // höchstes bit von lsb sagt aus, ob 0,5 Grad dazu addiert werden sollen
  lsb = (lsb & 0x80 ) >> 7; // nun ist lsb = 0 oder 1
  Wire.endTransmission();
  if (msb1 < 0x80) { // Positiver Wert?
    return double(msb + lsb) / 2; // positiver Wert
  }
  else {
    return double(msb + lsb) / 2 - 128; // negativer Wert
  }
}

// LM75 Konfigurationsregister setzen, Werte wie oben definiert
void set_LM75_config(int device, byte value)
{
  Wire.beginTransmission(SensorAdresse + device);
  Wire.write(1); // Select Konfigurationsregister
  Wire.write(value);
  Wire.endTransmission();
}

// LM75 Konfigurationsregister auslesen, device = 0-7
byte get_LM75_config(int device)
{
  byte reg;
  Wire.beginTransmission(SensorAdresse + device);
  Wire.write(1); // Select Konfigurationsregister
  Wire.endTransmission();
  Wire.requestFrom(SensorAdresse + device, 1);
  if (Wire.available()) {
    reg = Wire.read();
  }
  Wire.endTransmission();
  return reg;
}

// LM75 Schaltwerte setzen, device = 0-7, regx = Wert, Grad als double
void set_LM75_schaltwert(int device, byte regx, double grad)
{
  int8_t msb;
  int8_t lsb = 0;
  uint8_t y = 0;
  boolean neg = false;
  if (grad < 0) {
    msb = abs(int(grad)) + 128;
  }
  else {
    msb = abs(int(grad));
  }
  if (grad - abs(int(grad)) > 0) {
    lsb = 0x80;
  }
  Wire.beginTransmission(SensorAdresse + device);
  Wire.write(regx); // Selektiere oberes oder unteres Register
  Wire.write(msb);
  Wire.write(lsb);
  Wire.endTransmission();
}
